#! /usr/bin/env python


from .scene_base import Scene_Base

class Scene_Insteon(Scene_Base):

    def __init__(self, container, scene_info):
        Scene_Base.__init__(self,container, scene_info)

        self.add_property('onoff')

    def process_websocket_event(self,event):
        pass # there are no events for scenes AFAICT

    def turn_on(self):
        path = ('nodes/' + self.address + '/cmd/DON')
        return self.send_request(path)

    def turn_off(self):
        path = ('nodes/' + self.address + '/cmd/DOF')
        return self.send_request(path)

    def device_event(self,device): #device event, process and see if we are interested
        if device.family == '1': #insteon device
            if device.address in self.responders: # this scene has this device
                scene_onoff = 'off'
                for address in self.responders:
                    device = self.container.get_device(address)
                    if device is not None:
                        if device.category == '1': #insteon dimmer
                            if device.get_property ('level') > 0:
                                scene_onoff ='on'
                        if device.category == '2': #insteon switch
                            if device.get_property ('onoff') == 'on':
                                scene_onoff ='on'
                
                if scene_onoff != self.properties ['onoff']:
                    self.set_property ('onoff',scene_onoff)




#! /usr/bin/env python

''' 

Base Device

Common to all elements returned from rest/nodes/devices

device_type = switch,dimmer,contact

'''


from .. item_base import Item_Base


class Device_Base(Item_Base):

    def __init__(self, container, device_type, name):
        Item_Base.__init__(self,container, name)

        self.device_type = device_type





#!/usr/local/bin/python
import socket
import struct
from threading import Thread, Timer

TIMEOUT = 1.5

class Discover (object):

    def __init__(self):
        pass

    def start(self):
        ISYs = []
        MCAST_GRP = '239.255.255.250'
        MCAST_PORT = 1900
        msg = "M-SEARCH * HTTP/1.1\r\nHOST:239.255.255.250:1900\r\nMAN:\"ssdp.discover\"\r\nMX:1\r\nST:urn:udi-com:device:X_Insteon_Lighting_Device:1\r\n\r\n".encode()
        
        self.discovery_done = False

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 5)
        mreq = struct.pack('4sl', socket.inet_aton(MCAST_GRP), socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        sock.settimeout(0.1)

        def multicast_timeout():
            self.discovery_done = True

        timer = Timer(TIMEOUT, multicast_timeout)
        timer.start()		
        sock.sendto(msg, (MCAST_GRP, MCAST_PORT))
        
        while self.discovery_done != True:
            try:
                data = sock.recv(10240).decode()

                if 'HTTP/1.1 200 OK' in data:
                    index = data.index('LOCATION:')
                    index2 = data.index('/desc')
                    location = data[index+16:index2]
                    index = data.index('USN:')
                    index2 = data.index('::urn')
                    uuid = data[index+4:index2]
                    ISY = '%s [%s]' % (location, uuid)
                    #print('isy found',ISY)
                    ISYs.append(location)
            except Exception as e: 
                pass

        sock.close()
        return ISYs



if __name__ == "__main__":
    d = Discover()
    isys = d.start()

    print(isys)

    for x in isys:
        print(x)


#! /usr/bin/env python

''' 

returns a device instance using node data from an insteon device, None if unable to create device

'''

from .device_insteon_dimmer import Device_Insteon_Dimmer
from .device_insteon_switch import Device_Insteon_Switch
from .device_insteon_fan import Device_Insteon_Fan

insteon_device_classes = {
    '1' : Device_Insteon_Dimmer,
    '2' : Device_Insteon_Switch,
}

'''
dev_cat_sub_cat = {
    '1' : {
        '46' : Device_Insteon_Fan,
    },
}

'''
def get_insteon_device_class (device_info):

    #TBD add support for other device families
    if device_info.family == '1': #insteon devices, sanity check
        #look for specific device dev/sub cat that need special handling
        #print(device_info)
        if device_info.category == '1' and device_info.sub_category == '46' and device_info.address_parts [3] == '2': # fanlinc motor
            return Device_Insteon_Fan

        #if device_info.category in dev_cat_sub_cat and device_info.sub_category in dev_cat_sub_cat [device_info.category]:

        #override device cat for keypadlinc dimmer buttons and change to switch type devices
        if device_info.node_def_id.find('KeypadButton') and device_info.address_parts [3] != '1': # maybe use node flag
            device_info.category = '2'

        #find device class
        #check for specific devcat/subcat

        if device_info.category in insteon_device_classes:
            device_class = insteon_device_classes [device_info.category]
            return device_class
            


#! /usr/bin/env python

import time
import datetime
import traceback
import threading 

from .items.devices.device_container import Device_Container 
from .items.scenes.scene_container import Scene_Container 
from .items.variables.variable_container import Variable_Container 
from .items.programs.program_container import Program_Container 
from .items.controller.controller_container import Controller_Container 

from .network.http_client import HTTP_Client
from .network.websocket_client import Websocket_Client
from .network.discover import Discover

import logging
logger = logging.getLogger(__name__)

#event categores controller, device, scene, variable, program


class Controller(object):

    def __init__(self,address=None,port=None,username='admin',password='admin',use_https=False,event_handler=None):
        if address == None:
            discover = Discover()
            controller_list = discover.start()
            if len(controller_list) > 0:
                address = controller_list[0]

        if address is None:
            logger.error('No controller address found')
            quit()

        self.address = address
        self.port = port
        self.username = username
        self.password = password

        self.event_handlers = []
        if event_handler is not None:
            self.event_handlers.append(event_handler)

        self.http_client = HTTP_Client (address,port,username,password,use_https)

        self.controller_container = Controller_Container(self)
        self.device_container = Device_Container(self)
        self.scene_container = Scene_Container(self)
        self.variable_container = Variable_Container(self)
        self.program_container = Program_Container(self)

        self.last_heartbeat = None

        self.controller_container.start()        
        
        self.start()


    def start(self):
        success = True
        self.process_controller_event('status','init')        

        self.device_container.start()
        if self.device_container.items_retrieved is False:
            success = False

        self.scene_container.start() 
        if self.scene_container.items_retrieved is False:
            success = False  

        self.variable_container.start() 
        if self.variable_container.items_retrieved is False:
            success = False 

        self.program_container.start()
        if self.program_container.items_retrieved is False:
            success = False
        
        if success:
            self.process_controller_event('status','ready')
            self.websocket_client = Websocket_Client(self,self.address,self.port,self.username,self.password,False)
        else:
            self.process_controller_event('status','error')
            self.retry_start(10)

    def retry_start(self,delay_seconds):
        def restart ():
            self.start()

        self.restart_timer = threading.Timer(delay_seconds, restart) 
        self.restart_timer.start()
        
    def container_event(self,container,item,event,*args):
        #print ('Event {} from .{}: {} {}'.format(item.name,container.container_type,item,args))
        self.publish_container_event(container,item,event,*args)

    def publish_container_event(self,container,item,event,*args):
        for event_handler in self.event_handlers:
            try:
                event_handler (container,item,event,args)
            except Exception as ex:
                logger.error('Event handler Error {}'.format(ex))
                traceback.print_exc()
            
    def send_request(self,path,query=None,timeout=None): 
        success,response = self.http_client.request(path,query,timeout)
        if success:
            self.process_controller_event('http','connected')
        else:
            self.process_controller_event('http','error')
        return success, response

    def websocket_connected(self,connected): #True websocket connected, False, no connection
        if connected:
            self.process_controller_event('websocket','connected')
        else:
            self.process_controller_event('websocket','disconnected')

    def websocket_event(self,event): #process websocket event
        #print ('WS Event {}'.format(event))
        try:

            if event.address is not None: #event from .a device/node
                self.device_container.websocket_event (event)

            if event.control == '_0': # heartbeat
                self.process_heartbeat(event)

            elif event.control == '_1': # trigger
                if event.action == '0': #program
                    self.program_container.websocket_event (event)
                elif event.action == '6': # variable change
                    self.variable_container.websocket_event (event)

            elif event.control == '_5': # system status
                self.process_system_status(event)

        except Exception as ex:
                logger.error('websocket handler Error {}'.format(ex))
                traceback.print_exc()

    def process_controller_event(self,property_,value):
        controller = self.controller_container.get('controller')
        controller.set_property(property_,value)

    def process_heartbeat(self,event):
        self.last_heartbeat = datetime.datetime.now()
        self.process_controller_event('heartbeat',self.last_heartbeat)

    def process_system_status(self,event):
        if event.action =='0': #idle
            self.process_controller_event('state','idle')
        elif event.action == '1':#busy
            self.process_controller_event('state','busy')





#! /usr/bin/env python

''' 

'''
from .. item_base import Item_Base

class Controller_Base(Item_Base):

    def __init__(self, container, id):
        Item_Base.__init__(self,container,'ISY Controller')

        self.id = id
        self.name = id

        self.add_property('state', 'idle') 
        self.add_property('heartbeat','none')
        self.add_property('websocket','disconnnected')
        self.add_property('http','disconnected')

    def __str__(self):
        return ("Controller: {}; state {}; status {}".format(self.id, self.properties['state'], self.properties['status']))

 


#! /usr/bin/env python

''' 

Base Device

Common to all elements returned from rest/nodes/devices

   properties = list of properties and values

'''



class Device_Insteon_Base(object):

    def __init__(self, device_info):
        self.family = device_info.family
        self.category = device_info.category
        self.sub_category = device_info.sub_category
        self.version = device_info.version
        self.res = device_info.reserved
        self.address = device_info.address
        self.address_parts = device_info.address_parts
        self.flag = device_info.flag
        #self.container_node_address = device_info.container_node_address

    @property
    def flag(self):
        return self._flag

    @flag.setter
    def flag(self,flag):
        self._flag = flag
        if flag & 16:
            self.set_property ('status','alert')
        else:
            self.set_property ('status','ready')
        



#! /usr/bin/env python


from .device_speed_controller import Device_Speed_Controller
from .device_insteon_base import Device_Insteon_Base

speeds=['off','low','medium','high']

class Device_Insteon_Fan(Device_Speed_Controller,Device_Insteon_Base):

    def __init__(self, container, device_info):
        Device_Speed_Controller.__init__(self,container,device_info.name,speeds)
        Device_Insteon_Base.__init__(self,device_info)

        if device_info.property_value:
            self.set_property ('speed',self.level_to_speed(int(device_info.property_value)))

    def process_websocket_event(self,event):
            if event.control == 'ST':
                self.set_property ('speed',self.level_to_speed(int(event.action)))

    def set_speed (self,speed):
        level = self.speed_to_level(speed)
        path = ('nodes/' + self.address + '/cmd/DON/' + str(level))
        return self.send_request(path)
 
    def level_to_speed(self,level):
        if level == 0:
            return'off'
        elif level > 0 and level <= 63:
            return 'low'
        elif level >= 50 and level <= 191:
           return 'medium'
        else:
            return 'high'

    def speed_to_level(self,speed):
        if speed == 'off':
            return 0
        elif speed == 'low':
            return 63
        elif speed == 'medium':
            return 99
        else:
            return 255






#! /usr/bin/env python


from .device_base import Device_Base


class Device_Speed_Controller(Device_Base):

    def __init__(self, container, name ,speeds=['off','low','medium','high']):
        Device_Base.__init__(self,container,'speed_controller',name)

        self.speeds = speeds

        self.add_property('speed','unknown') 

    def set_speed(self,speed):
        pass




#! /usr/bin/env python


# process a node and assemble the info to build a program


class Program_Info(object):

    def __init__(self,node):

        self.node = node

        self.valid = False

        try:
            if node.attrib ['folder'] == 'true': # program folder, not interested
                return

            if node.attrib ['enabled'] == 'false': # not enabled
                return

            self.id = node.attrib ['id']
            self.name = node.find('name').text
            self.last_run_time = node.find('lastRunTime').text
            self.last_finish_time = node.find('lastFinishTime').text

            self.valid = True
        except:
            pass

    def __repr__(self):
        return 'Program: Name {} ID {}'.format(self.name,self.id)
    



import time

from ISY.controller import Controller

url = '192.168.1.213'

def print_events(container,item,event,*args):
    print ('Event {} from {}: {} {}'.format(event,container.container_type,item.name,args))


try:
    c = Controller(url,username='admin',password='admin',use_https=False,event_handler=print_events)
    time.sleep(2)  
    #device = c.device_container.get('14 A9 92 2')
    #device = c.device_container.get('42 C8 99 1')
    #print ('got device',device)

    #scene = c.scene_Container.get_scene('25770')
    #print ('got scene',scene)

    #program = c.program_Container.get_program ('0022')

    while True:
        time.sleep(2)
        #device.set_level (0)
        #device.set_speed ('low')
        #device.set_speed ('medium')
        #device.set_speed ('high')
        #scene.turn_on()
        #program.run()
        time.sleep(2)
        #device.set_speed ('off')
        #scene.turn_off()
        #device.set_level (100)
        #program.run_else()

except KeyboardInterrupt:
    print("KeyboardInterrupt has been caught.")



#! /usr/bin/env python

import xml.etree.ElementTree as ET
import traceback

from ..item_container import Item_Container

from .device_info import Device_Info

from .insteon_device_manager import get_insteon_device_class



import logging
logger = logging.getLogger(__name__)


class Device_Container (Item_Container):

    def __init__(self, controller):
        Item_Container.__init__(self,controller,'Device')

    def start(self):
        success,response = self.send_request('nodes/devices')

        if success and response.status_code == 200:
            try:
                root = ET.fromstring (response.content)        
                self.process_device_nodes (root)    
                self.items_retrieved = True
            except Exception as ex:
                    logger.error('container error {}'.format(ex))
                    traceback.print_exc()

    def process_device_nodes(self,root):
        for device in root.iter('node'):
            self.process_device_node(device)
                
    def process_device_node(self,node):
        device_info = Device_Info(node) # parse node XML
        #print('process node',device_info.valid,device_info)

        if device_info.valid: # make sure we have the info we need
            ''' device family
            	0 = Default (core driver implementation e.g. Insteon, UPB)
				1 = Insteon products
				2 = UPB Products
				3 = RCS Products
				4 = ZWave products
				5 = Auto DR (for groups)
				6 = Generic (for groups)
				7 = UDI Products
				8 = Brultech Products
				9 = NCD Products
            '''
            if device_info.family == '1': #insteon devices
                device_class = get_insteon_device_class (device_info)
                if device_class is not None:
                    device = device_class(self,device_info)
                    self.add(device,device.address)

            
         
    def websocket_event(self,event):
        #print('Device event',event)
        device = self.get(event.address)
        if device is not None:
            device.process_websocket_event(event)

    
    
        




#! /usr/bin/env python


import xml.etree.ElementTree as ET
import traceback

from .. item_container import Item_Container
from .scene_info import Scene_Info
from .scene_insteon import Scene_Insteon

import logging
logger = logging.getLogger(__name__)


scene_classes = {
    '6' : Scene_Insteon,
}

class Scene_Container (Item_Container):

    def __init__(self, controller):
        Item_Container.__init__(self,controller,'Scene')

    def start(self):
        success,response = self.controller.send_request('nodes/scenes')

        try:
            if success and response.status_code == 200:
                root = ET.fromstring (response.content)        
                self.process_scene_nodes (root)       
                self.items_retrieved = True

        except Exception as ex:
            logger.error('container manager Error {}'.format(ex))
            traceback.print_exc()

    def process_scene_nodes(self,root):
        for scene in root.iter('group'):
            self.process_scene_node(scene)
                
    def process_scene_node(self,node):
        scene_info = Scene_Info(node)

        if scene_info.valid: # make sure we have the info we need
            #print('process scene',scene_info)
            if scene_info.family in scene_classes:
                scene_class = scene_classes [scene_info.family]

                scene = scene_class(self,scene_info)
                self.add(scene,scene.id)
        else:
            logger.warn ('Invalid scene info {}'.format(scene_info.name))
         
    def device_event(self,device): # notification from controller about a device event, used to "track" scene state
        for address,scene in self.list.items():
            scene.device_event (device)

    def get_device(self,address):
        return self.controller.device_Container.get_device(address)
    
        




#! /usr/bin/env python


import xml.etree.ElementTree as ET
import traceback

from .. item_container import Item_Container
from .program_info import Program_Info 
from .program_base import Program_Base


import logging
logger = logging.getLogger(__name__)


class Program_Container (Item_Container):

    def __init__(self, controller):
        Item_Container.__init__(self,controller,'Program')

    def start(self):
        success,response = self.controller.send_request('programs','subfolders=true')

        if success and response.status_code == 200:
            try:
                root = ET.fromstring (response.content)        
                self.process_program_nodes (root)    
                self.items_retrieved = True
            except Exception as ex:
                logger.error('container error {}'.format(ex))
                traceback.print_exc()

    def process_program_nodes(self,root):
        for program in root.iter('program'):
            self.process_program_node(program)
                
    def process_program_node(self,node):
        program_info = Program_Info(node) # parse node XML

        if program_info.valid: # make sure we have the info we need
            program = Program_Base(self,program_info)
            self.add(program,program.id)
         
    def websocket_event(self,event):
        #print('Program event',event)
        id = event.event_info_node.find('id').text.zfill(4) 

        status = None
        run_time = None
        finish_time = None
        
        if event.event_info_node.find('s') is not None:
            status = int(event.event_info_node.find('s').text )

        if event.event_info_node.find('r') is not None:
            run_time = event.event_info_node.find('r').text 

        if event.event_info_node.find('f') is not None:
            finish_time = event.event_info_node.find('f').text 

        if status and run_time and finish_time:
            program = self.get(id)
            if program:
                program.process_websocket_event (status,run_time,finish_time)
            else:
                logger.warning('Unable able to find program type id {}'.format(id))

